unit MoteurP4;

interface

uses Types, extCtrls, Dialogs, SysUtils, Grids;

const
     // incrmentation dans les 8 DIRECTIONS, regroupes 2 par 2 pour 4 SENS
     DIR_INCR : array[ 0..7] of TPoint = ( ( x:1;y:0),( x:-1;y:0),    // horizontale
                                           ( x:0;y:1),( x:0;y:-1),    // vertical
                                           ( x:1;y:1),( x:-1;y:-1),   // oblique slash
                                           ( x:-1;y:1),( x:1;y:-1) ); // oblique anti-slash

// CLASSE JOUEUR
type
  TPlateau = class;
  TJoueur = class( TObject )
  protected
    Plateau: TPlateau; // plateau propritaire
    ICamp: Integer; // camps... peut tre diffrent du numro du joueur
    SNom: string; // son nom
    IJetonsRestants : Integer; // nombre de jetons restants  jouer
  public
    // gestion
    constructor Create( zPlateau: TPlateau );
    destructor Destroy; override;
    procedure SetNom( zSNom: string ); // dfinition du nom du joueur
    // actions
    procedure Init( zSNom: string; zICamp: Integer ); // init gnral du joueur, son camp
    function Jouer( ICol: Integer ): Boolean; // Le joueur joue dans la colonne spcifie
  end;

// CLASSE PLATEAU
  TPlateau = class( TObject )
  protected
    // graphique
    Conteneur: TPanel;
    StringGrid: TStringGrid;
    // Plateau
    Grille: array[ 0..6, 0..5 ] of Integer;
    DernierJeton: TPoint;
    // Utilitaires
    function CaseVide( ICol: Integer ): Integer; // renvoie n de ligne dispo la plus basse
    function ResultatDernierJeton: Integer; // -1 nul, 0 en cours, 1 ou 2 gain
  public
    // Accs publique
    ICurseur: Integer; // colonne du jeton  jouer 0..6
    IJoueur: Integer; // 0: aucun, 1,2: numro du joueur
    Joueur: array[ 1..2 ] of TJoueur; // les 2 joueurs
    // gestion
    constructor Create( zConteneur: TPanel; zStringGrid: TStringGrid );
    destructor Destroy; override;
    // Actions
    procedure NouveauJeu( SNom1, SNom2: string; IPreums: Integer ); // lance une nouvelle partie
    procedure Jouer; // Joue un coup selon l'tat courant
    procedure Fin( SMessage: string ); // Fin de la partie + message
    // Interaction graphiques
    procedure PionAfficher( ICamp: Integer; PPos: TPoint ); //afficher un jeton en coord logique
    procedure PionGlisser( ICamp, ICol, ILigneDest: Integer ); // chute d'une jeton
  end;


implementation

{ ============================================================================
  TJOUEUR
  ============================================================================ }
constructor TJoueur.Create( zPlateau: TPlateau );
begin
     Plateau := zPlateau;
     ICamp := 0;
     IJetonsRestants := 0;
end;

destructor TJoueur.Destroy;
begin
     //
     inherited;
end;

procedure TJoueur.SetNom( zSNom: string );
begin
     if zSNom <> ''
       then SNom := zSNom;
end;

procedure TJoueur.Init( zSNom: string; zICamp: Integer );
begin
     SetNom( zSNom );
     ICamp := zICamp;
     IJetonsRestants := 21;
end;

function TJoueur.Jouer( ICol: Integer ): Boolean;
var
   ILigne: Integer;
begin
     Result := False;
     ILigne := Plateau.CaseVide( ICol );
     if ILigne <> -1 then
       begin
       Plateau.Grille[ ICol, ILigne ] := ICamp;
       Plateau.DernierJeton := Point( ICol, ILigne );
       Dec( IJetonsRestants );
       Result := True;
       end;
end;

{ ============================================================================
  TPLATEAU
  ============================================================================ }
constructor TPlateau.Create( zConteneur: TPanel; zStringGrid: TStringGrid );
begin
     // var
     Conteneur := zConteneur;
     Conteneur.DoubleBuffered := True;
     StringGrid := zStringGrid;
     StringGrid.Selection := TGridRect( Rect( -1, -1, -1, -1 ) );
     Joueur[ 1 ] := TJoueur.Create( Self );
     Joueur[ 2 ] := TJoueur.Create( Self );
     ICurseur := 3;
end;

destructor TPlateau.Destroy;
begin
     Joueur[ 1 ].Free;
     Joueur[ 2 ].Free;
     inherited;
end;

procedure TPlateau.NouveauJeu( SNom1, SNom2: string; IPreums: Integer );
var
   x, y: Integer;
begin
     // Raz de la grille logique
     for x := 0 to 6 do
       for y := 0 to 5
         do Grille[ x, y ] := 0;
     // Raz de la grille physique
     for x := 0 to 6 do
       for y := 0 to 6
         do StringGrid.Cells[ x, y ] := '';
     // Autres inits
     Joueur[ 1 ].Init( SNom1, 1 );
     Joueur[ 2 ].Init( SNom2, 2 );
     IJoueur := IPreums;
end;

procedure TPlateau.Jouer;
var
   IEtat: Integer;
begin
     // Fait jouer le joueur actuel
     if Joueur[ IJoueur ].Jouer( ICurseur ) then
       begin
       PionGlisser( Joueur[ IJoueur ].ICamp, DernierJeton.X, DernierJeton.Y );
       IEtat := ResultatDernierJeton;
       // Rsultat du coup - -1 / 0 / 1 ou 2
       case IEtat of
         -1  : Fin( 'PARTIE NULLE !' );
         1,2 : Fin( Joueur[ IEtat ].SNom + #13 + #10 + 'is the winner !' );
         else  IJoueur := ( IJoueur mod 2 ) + 1;
       end;
       end;

end;

procedure TPlateau.Fin( SMessage: string );
begin
     IJoueur := 0;
     ShowMessage( SMessage );
end;

function TPlateau.CaseVide( ICol: Integer ): Integer;
var
   i: Integer;
begin
     Result := -1;
     for i := 0 to 5 do
       if Grille[ ICol, i ] = 0
         then Result := i
         else Break;
end;

function TPlateau.ResultatDernierJeton: Integer;
var
   i, j: Integer;
   TabDirCompteur: array[ 0..7 ] of Integer; // Compteur par direction
   TabDirOK: array[ 0..7 ] of Boolean; // Directions  exploiter, False: ne plus exploiter
   xi, yi: Integer;
begin
     // Inits
     Result := 0;
     for i := 0 to 7 do
       begin
       TabDirCompteur[ i ] := 0; // 0 jeton conscutif
       TabDirOK[ i ] := True; // direction exploitable
       end;

     // Partie Gagne ? - on regarde autour du dernier jeton
     for i := 0 to 7 do // les 8 directions = 8 jetons potentiels autour
       for j := 1 to 3 do // 3 niveaux conscutifs + jeton central = test de 4 jetons aligns potentiels
         if TabDirOK[ i ] then // Direction exploitable ?
           begin
           // coord du jeton voisin
           xi := DernierJeton.X + DIR_INCR[ i ].X * j;
           yi := DernierJeton.Y + DIR_INCR[ i ].Y * j;
           if ( xi < 0 ) or ( xi > 6 ) or ( yi < 0 ) or ( yi > 5 ) then Continue; // hors du jeu
           // Cas du jeton voisin = jeton central -> incrmentation du sens en cours
           if Grille[ xi, yi ] = Grille[ DernierJeton.X, DernierJeton.Y ]
             then Inc( TabDirCompteur[ i ] )
             else TabDirOK[ i ] := False; // au 1er jeton diffrent, la direction n'est plus  exploiter
           end;

     // Bilan des compteurs, dans les 4 sens, 2 par 2 directions selon DIR_INCR 
     for i := 0 to 3 do
       if TabDirCompteur[ i*2 ] + TabDirCompteur[ i*2+1 ] >= 3 // il peut y avoir plus de 4 jetons aligns...
         then Result := IJoueur;
         
     // Partie Nulle ?
     if Result = 0 then
       if ( Joueur[ 1 ].IJetonsRestants = 0 ) and ( Joueur[ 2 ].IJetonsRestants = 0 )
         then Result := -1;
end;

// Afichage (ou effacement) du jeton dans la StringGrid
procedure TPlateau.PionAfficher( ICamp: Integer; PPos: TPoint );
var
   CCar: string;
begin
     case ICamp of
       1 : CCar := ' X';
       2 : CCar := ' O';
       else CCar := '';
     end;
     // le tableau logique commence en indice -1 pour le lacher de jeton
     StringGrid.Cells[ PPos.X, PPos.Y + 1 ] := CCar;
end;

procedure TPlateau.PionGlisser( ICamp, ICol, ILigneDest: Integer );
var
   y: Integer;
begin
     // dbut
     PionAfficher( ICamp, Point( ICol, -1 ) );
     Conteneur.Refresh;
     Sleep( 100 );
     // Glissement
     for y := -1 to ILigneDest - 1 do
       begin
       Sleep( 100 );
       PionAfficher( 0, Point( ICol, y ) );
       PionAfficher( ICamp, Point( ICol, y + 1 ) );
       Conteneur.Refresh;
       end;
end;



end.
